<!DOCTYPE html>
<html lang="en" ng-app="AgoraApp" style="height:100%">
<head>
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Agora</title>
    <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.11.3/jquery.min.js"></script>
    <script src="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.6/js/bootstrap.min.js"
            integrity="sha384-0mSbJDEHialfmuBBQP6A4Qrprq5OVfW37PRR3j5ELqxss1yVqOtnepnHVP9aJ7xS"
            crossorigin="anonymous"></script>
    <link rel="stylesheet"
          href="https://fonts.googleapis.com/css?family=Roboto|Roboto:light">
    <link href="//netdna.bootstrapcdn.com/bootstrap/3.3.6/css/bootstrap.min.css" rel="stylesheet" media="screen">
    <link rel="icon" type="image/png" href="{{ url_for('static', filename='favicon.ico') }}">
    <link rel="stylesheet" href="https://unpkg.com/leaflet@1.1.0/dist/leaflet.css"
          integrity="sha512-wcw6ts8Anuw10Mzh9Ytw4pylW8+NAD4ch3lqm9lzAsTxg0GFeJgoAtxuCLREZSC5lUXdVyo/7yfsqFjQ4S+aKw=="
          crossorigin=""/>
    <link href='//cdn.jsdelivr.net/yasqe/2.11.10/yasqe.min.css' rel='stylesheet' type='text/css'/>
    <link href="{{ url_for('static', filename='styles.css') }}" rel="stylesheet" media="screen">
    <script src='//cdn.jsdelivr.net/yasqe/2.11.10/yasqe.bundled.min.js'></script>
    <script src="https://unpkg.com/leaflet@1.1.0/dist/leaflet.js"
            integrity="sha512-mNqn2Wg7tSToJhvHcqfzLMU6J4mkOImSPTxVZAdo+lcPlk+GhZmYgACEe0x35K7YzW1zJ7XyJV/TT1MrdXvMcA=="
            crossorigin=""></script>
    <script src="https://unpkg.com/terraformer@1.0.8"></script>
    <script src="https://unpkg.com/terraformer-wkt-parser@1.1.2"></script>
    <script src="https://ajax.googleapis.com/ajax/libs/angularjs/1.5.5/angular.min.js"></script>
    <script src="https://ajax.googleapis.com/ajax/libs/angularjs/1.5.5/angular-route.js"></script>
    <script src="https://ajax.googleapis.com/ajax/libs/angularjs/1.5.5/angular-animate.js"></script>
    <!--<script src="{{ url_for('static', filename='smart-table.js') }}"></script>-->
    <script src="{{ url_for('static', filename='oboe-browser.js') }}"></script>
    <script src="{{ url_for('static', filename='main.js') }}"></script>
    <base href="/">
</head>
<body style="height:100%">
<div style="position:absolute;z-index:9; width:100%; top:0">
    <nav class="navbar navbar-default" style="background-color: white; border: none">
        <div class="container-fluid">
            <!-- Brand and toggle get grouped for better mobile display -->
            <div class="navbar-header" style="width: 100%">
                <button type="button" class="navbar-toggle collapsed" data-toggle="collapse"
                        data-target="#bs-example-navbar-collapse-1" aria-expanded="false">
                </button>
                <a href="/" style="display:block ; float:left">
                    <img id="menu-toggle" class="navbar-brand"
                         src="{{ url_for('static', filename='agora-im.png') }}">
                </a>
                <div id="agora">Agora</div>
                <img src="{{ url_for('static', filename='upm.png') }}" height="40"
                     style="margin-top: 5px; margin-left:15px" align="right">
                <img src="{{ url_for('static', filename='logoOEG.png') }}" height="40" style="margin-top: 5px"
                     align="right">
            </div>

            <!-- Collect the nav links, forms, and other content for toggling -->
            <div class="collapse navbar-collapse" id="bs-example-navbar-collapse-1">
                <ul class="nav navbar-nav">
                </ul>
            </div><!-- /.navbar-collapse -->
        </div><!-- /.container-fluid -->
    </nav>
</div>
<div style="height:100%" ng-view>
</div>
</body>
</html>